#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Defines index mappings."""


def up_index(index):
    """Function to return up-orbital index given a spatial orbital index.

    Args:
        index (int): spatial orbital index

    Returns:
        An integer representing the index of the associated spin-up orbital
    """
    return 2 * index


def down_index(index):
    """Function to return down-orbital index given a spatial orbital index.

    Args:
        index (int): spatial orbital index

    Returns:
        An integer representing the index of the associated spin-down orbital
    """
    return 2 * index + 1
